--[[--------------------------------------------------------------------
	OnScreenHealth
	Minimal text-based HUD for the player, pet, and target units.
	by Phanx < addons@phanx.net >
	Copyright  20072010 Phanx. Some rights reserved. See LICENSE.txt for details.
	http://www.wowinterface.com/downloads/info7470-OnScreenHealth.html
	http://wow.curse.com/downloads/wow-addons/details/onscreenhealth.aspx
----------------------------------------------------------------------]]

local OnScreenHealth = CreateFrame("Frame", "OnScreenHealth")
OnScreenHealth:SetScript("OnEvent", function(f, e, ...) return f[e] and f[e](f, ...) end)
OnScreenHealth:RegisterEvent("ADDON_LOADED")

local ADDON_NAME, ns = ...
ns.OnScreenHealth = OnScreenHealth

local db, currentToT
local playerUnit = "player"
local playerClass = select(2, UnitClass("player"))

local unitFrames = { }
ns.unitFrames = unitFrames

local SharedMedia = LibStub("LibSharedMedia-3.0")

local classColors = { }
for class, color in pairs(RAID_CLASS_COLORS) do
	classColors[class] = ("%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end

local units = {
	["pet"]     = true,
	["player"]  = true,
	["target"]  = true,
	["vehicle"] = true,
}

------------------------------------------------------------------------

if not ns.L then
	ns.L = { }
end

local L = setmetatable(ns.L, {
	__index = function(t, k)
		t[k] = k
		return k
	end
})

L["Mana"] = MANA
L["Rage"] = RAGE
L["Energy"] = ENERGY
L["Focus"] = FOCUS
L["Runic Power"] = RUNIC_POWER
L["Holy Power"] = HOLY_POWER
L["Soul Shards"] = SOUL_SHARDS:gsub("|4.-:(.-);", "%1")

L["Maelstrom Weapon"] = GetSpellInfo(53817)
L["Shadow Orb"] = GetSpellInfo(77487)
L["Shadow Orbs"] = GetSpellInfo(77486)

------------------------------------------------------------------------

local function Debug(lvl, str, ...)
	if lvl > 0 then return end
	if (...) then
		if str:match("%%") then
			str = str:format(...)
		else
			str = string.join(", ", str, tostringall(...))
		end
	end
	return print("|cffff6666[DEBUG] OnScreenHealth:|r", str)
end
ns.Debug = Debug

------------------------------------------------------------------------

local function si(v)
	if not v then return "" end
	if type(v) ~= "number" then return v end
	Debug(5, "si(%d)", v)

	if v >= 10000000 then
		return ("%.1fm"):format(v / 1000000)
	elseif v >= 1000000 then
		return ("%.2fm"):format(v / 1000000)
	elseif v >= 100000 then
		return ("%.0fk"):format(v / 1000)
	elseif v >= 10000 then
		return ("%.1fk"):format(v / 1000)
	else
		return v
	end
end

------------------------------------------------------------------------

local function GetUnitColor(unit)
	if not unit then return end

	if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) then
		return db.colorAbsent
	else
		local color = UnitIsPlayer(unit) and classColors[select(2, UnitClass(unit))] or nil
		if color then
			return color
		else
			local r, g, b = GameTooltip_UnitColor(unit)
			return string.format("%02x%02x%02x", r * 255, g * 255, b * 255)
		end
	end
end

------------------------------------------------------------------------

local function GetFormattedText(text, cur, max)
	if not text and cur and max then return end
	Debug(3, "GetFormattedText")

	text = text:gsub("$c:s", si(cur))
	text = text:gsub("$c", cur)

	text = text:gsub("$m:s", si(max))
	text = text:gsub("$m", max)

	text = text:gsub("$d:s", si(max - cur))
	text = text:gsub("$d", max - cur)

	text = text:gsub("$p", floor(cur / max * 100))

	return text
end

------------------------------------------------------------------------

function OnScreenHealth:UNIT_HEALTH(unit)
	if not units[unit] then return end
	Debug(2, "Updating health for %s", unit)

	local f = unitFrames[unit]
	if not f or not f:IsShown() then return end

	if not UnitIsConnected(unit) then
		return f.Health:SetFormattedText("|cff%s%s|r", db.colorAbsent, L["Offline"])
	elseif UnitIsGhost(unit) then
		return f.Health:SetFormattedText("|cff%s%s|r", db.colorAbsent, L["Ghost"])
	elseif UnitIsDead(unit) then
		return f.Health:SetFormattedText("|cff%s%s|r", db.colorAbsent, L["Dead"])
	end

	local cur, max = UnitHealth(unit), UnitHealthMax(unit)
	local percent = cur / max

	local color
	if db.colorHealthMode == "CLASS" then
		color = GetUnitColor(unit)
	elseif db.colorHealthMode == "GRADIENT" then
		local r1, g1, b1, r2, g2, b2
		if percent <= 0.5 then
			percent = percent * 2
			r1, g1, b1 = unpack(db.colorHealthMin)
			r2, g2, b2 = unpack(db.colorHealthMid)
		else
			percent = percent * 2 - 1
			r1, g1, b1 = unpack(db.colorHealthMid)
			r2, g2, b2 = unpack(db.colorHealthMax)
		end
		color = ("%02x%02x%02x"):format((r1 + (r2 - r1) * percent) * 255, (g1 + (g2 - g1) * percent) * 255, (b1 + (b2 - b1) * percent) * 255)
	else
		local r, g, b
		if percent < 0.33 then
			r, g, b = unpack(db.colorHealthMin)
		elseif percent < 0.67 then
			r, g, b = unpack(db.colorHealthMid)
		else
			r, g, b = unpack(db.colorHealthMax)
		end
		color = ("%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end

	f.Health:SetFormattedText("|cff%s%s|r", color, GetFormattedText(UnitIsPlayer(unit) and db.textFormatHealthPlayer or db.textFormatHealthMonster, cur, max))

	if db.alphaMode == "HEALTH" then
		if percent < 0.25 then
			f:SetAlpha(1)
		else
			f:SetAlpha(1.25 - percent)
		end
	end
end

OnScreenHealth.UNIT_MAXHEALTH = OnScreenHealth.UNIT_HEALTH

------------------------------------------------------------------------

local realPowerTypes = {
	["AMMOSLOT"] = true,
	["ENERGY"] = true,
	["FOCUS"] = true,
	["FUEL"] = true,
	["MANA"] = true,
	["RAGE"] = true,
	["RUNIC_POWER"] = true,
}
local resourceTypes = {
	["ECLIPSE"] = "DRUID",
	["HOLY_POWER"] = "PALADIN",
	["SOUL_SHARDS"] = "WARLOCK",
}
local powerTypeID = {
	["MANA"] = SPELL_POWER_MANA,
	["RAGE"] = SPELL_POWER_RAGE,
	["FOCUS"] = SPELL_POWER_FOCUS,
	["ENERGY"] = SPELL_POWER_ENERGY,
	["RUNIC_POWER"] = SPELL_POWER_RUNIC_POWER,
	["AMMOSLOT"] = SPELL_POWER_AMMOSLOT,
	["FUEL"] = SPELL_POWER_FUEL,
	["SOUL_SHARDS"] = SPELL_POWER_SOUL_SHARDS,
	["ECLIPSE"] = SPELL_POWER_ECLIPSE,
	["HOLY_POWER"] = SPELL_POWER_HOLY_POWER,
}

function OnScreenHealth:UNIT_POWER(unit, powerType)
	if not units[unit] then return end
	Debug(2, "UNIT_POWER", unit, powerType)

	if powerType == nil then
		powerType = select(2, UnitPowerType(unit))
	end

	if unit == "player" and resourceTypes[powerType] == playerClass then -- check class because Blizzard tells the client that everyone gains 3 holy power when they're resurrected -.-
		local amount = UnitPower(unit, powerTypeID[powerType])
		if powerType == "ECLIPSE" then
			local color
			if amount < 0 then
				color = db.powerColors.ECLIPSE_LUNAR
			elseif amount > 0 then
				color = db.powerColors.ECLIPSE_SOLAR
			else
				color = "ffffff"
			end
			local direction = GetEclipseDirection()
			if direction == "moon" then
				unitFrames.player.Resource:SetFormattedText("< |cff%s%d|r", color, abs(amount))
			elseif direction == "sun" then
				unitFrames.player.Resource:SetFormattedText("|cff%s%d|r >", color, abs(amount))
			else
				unitFrames.player.Resource:SetFormattedText("|cff%s%d|r", color, abs(amount))
			end
		elseif amount > 0 then
			unitFrames.player.Resource:SetFormattedText("|cff%s%d|r", db.powerColors[powerType], amount)
		else
			unitFrames.player.Resource:SetText()
		end
	end
	if not realPowerTypes[powerType] then
		-- what is this, I don't even
		return
	end

	local f = unitFrames[unit]
	if not f or not f:IsShown() then return end

	if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) or UnitManaMax(unit) == 0 then
		return f.Power:SetText()
	end

	local color = db.powerColors[powerType] or db.powerColors.MANA

	f.Power:SetFormattedText("|cff%s%s|r", color, GetFormattedText(UnitIsPlayer(unit) and db.textFormatPowerPlayer or db.textFormatPowerMonster, UnitMana(unit), UnitManaMax(unit)))
end

OnScreenHealth.UNIT_MAXPOWER = OnScreenHealth.UNIT_POWER
OnScreenHealth.UNIT_DISPLAYPOWER = OnScreenHealth.UNIT_POWER

------------------------------------------------------------------------

local resourceBuff = playerClass == "SHAMAN" and L["Maelstrom Weapon"] or playerClass == "PRIEST" and L["Shadow Orb"]
local resourceType = playerClass == "SHAMAN" and "MAELSTROM_WEAPON" or playerClass == "PRIEST" and "SHADOW_ORBS"

function OnScreenHealth:UNIT_AURA(unit)
	if unit ~= "player" or playerUnit ~= "player" or not unitFrames.player:IsShown() then return end

	local name, _, icon, count = UnitBuff("player", resourceBuff)
	if name then
		unitFrames.player.Resource:SetFormattedText("|cff%s%d|r", db.powerColors[resourceType], count)
	else
		unitFrames.player.Resource:SetText()
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UpdateAll(type)
	if type and type == "health" then
		self:UNIT_HEALTH(playerUnit)
		self:UNIT_HEALTH("target")
		self:UNIT_HEALTH("pet")
	elseif type and type == "power" and db.showPower then
		self:UNIT_POWER(playerUnit)
		self:UNIT_POWER("target")
		self:UNIT_POWER("pet")
		if (playerClass == "SHAMAN" and select(5, GetTalentInfo(2, 17)) > 0) or (playerClass == "PRIEST" and GetPrimaryTalentTree() == 3) then
			self:UNIT_AURA("player")
		end
	else
		self:UpdatePlayer()
		self:PLAYER_TARGET_CHANGED()
		self:UNIT_PET("player")
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UpdatePlayer()
	if not db.playerEnable or not (db.playerShowOOC or UnitAffectingCombat(playerUnit)) or not (db.playerShowDead or not UnitIsDeadOrGhost(playerUnit)) then
		unitFrames.player:Hide()
		return
	end
	Debug(1, "Updating player")

	unitFrames.player:Show()
	self:UNIT_HEALTH(playerUnit)
	if db.showPower then
		self:UNIT_POWER(playerUnit)
		if (playerClass == "SHAMAN" and select(5, GetTalentInfo(2, 17)) > 0) or (playerClass == "PRIEST" and GetPrimaryTalentTree() == 3) then
			self:UNIT_AURA("player")
		end
	end
	if db.playerShowCombatStatus then
		self:UpdateCombatStatus(playerUnit)
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UpdateCombatStatus(unit)
	if not unit then unit = playerUnit end
	Debug(1, "Updating "..unit.." status")

	if db[unit.."ShowCombatStatus"] and UnitAffectingCombat(unit) then
		unitFrames.player.Combat:SetText("|cff"..db.colorCombatStatus.."+|r")
	else
		unitFrames.player.Combat:SetText()
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UNIT_PET(owner)
	if owner and owner ~= "player" or db.disablePet then return end
	Debug(1, "UNIT_PET")

	if playerUnit == "vehicle" or not db.petEnable or not (db.playerShowOOC or UnitAffectingCombat(playerUnit)) or not (db.petShowDead or not UnitIsDeadOrGhost("pet")) or not UnitExists("pet") then
		unitFrames.pet:Hide()
		return
	end

	if UnitExists("pet") then
		if db.hidePetDead then
			if UnitIsDeadOrGhost("pet") then
				unitFrames.pet:Hide()
				return
			end
		end
		unitFrames.pet:Show()
		self:UNIT_HEALTH("pet")
		if db.showPower then
			self:UNIT_POWER("pet")
		end
	else
		unitFrames.pet:Hide()
	end
end

------------------------------------------------------------------------

function OnScreenHealth:PLAYER_TARGET_CHANGED()
	Debug(1, "PLAYER_TARGET_CHANGED")
	if not UnitExists("target") or not db.targetEnable or not (db.targetShowOOC or UnitAffectingCombat(playerUnit)) or not (db.targetShowDead or (not UnitIsDeadOrGhost("target") and UnitIsConnected("target"))) then
		unitFrames.target:Hide()
		currentToT = nil
		return
	end

	unitFrames.target:Show()
	self:UNIT_HEALTH("target")
	if db.showPower then
		self:UNIT_POWER("target")
	end
	self:UNIT_COMBO_POINTS()
end

------------------------------------------------------------------------

function OnScreenHealth:CheckToT()
	local newToT = UnitExists("targettarget") and UnitName("targettarget") or "NONE"
	if currentToT ~= newToT then
		currentToT = newToT
		OnScreenHealth:UpdateToT()
	end
end

function OnScreenHealth:UpdateToT()
	if UnitExists("target") and UnitExists("targettarget") then
		if UnitIsUnit(playerUnit, "targettarget") then
			unitFrames.target.Target:SetText("|cff"..db.colorTargetSelf..L["<<YOU>>"].."|r")
			return
		else
			unitFrames.target.Target:SetText("|cff"..GetUnitColor("targettarget")..UnitName("targettarget").."|r")
			return
		end
	else
		unitFrames.target.Target:SetText()
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UNIT_COMBO_POINTS()
	local cp = GetComboPoints(playerUnit, "target")
	if cp > 0 then
		unitFrames.target.ComboPoints:SetFormattedText("|cff%s%s|r", db.colorCombo, cp)
	else
		unitFrames.target.ComboPoints:SetText()
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UNIT_ENTERED_VEHICLE(unit)
	Debug(1, "UNIT_ENTERED_VEHICLE", unit)
	if unit ~= "player" then return end
	playerUnit = "vehicle"
	self:UNIT_HEALTH("vehicle")
	self:UNIT_POWER("vehicle")
	unitFrames.player.Resource:SetText()
	self:UNIT_PET("player")
end

function OnScreenHealth:UNIT_EXITED_VEHICLE(unit)
	Debug(1, "UNIT_EXITED_VEHICLE", unit)
	if unit ~= "player" then return end
	playerUnit = "player"
	self:UNIT_HEALTH("player")
	self:UNIT_POWER("player")
	if (playerClass == "SHAMAN" and select(5, GetTalentInfo(2, 17)) > 0) or (playerClass == "PRIEST" and GetPrimaryTalentTree() == 3) then
		self:UNIT_AURA("player")
	end
	self:UNIT_PET("player")
end

------------------------------------------------------------------------

function OnScreenHealth:PLAYER_REGEN_DISABLED()
	Debug(1, "PLAYER_REGEN_DISABLED")
	self:UpdatePlayer()
	self:UNIT_PET("player")
	self:PLAYER_TARGET_CHANGED()
	if db.alphaMode == "COMBAT" then
		local a = db.alphaCombat
		unitFrames.player:SetAlpha(a)
		unitFrames.target:SetAlpha(a)
		unitFrames.pet:SetAlpha(a)
	end
end

function OnScreenHealth:PLAYER_REGEN_ENABLED()
	Debug(1, "PLAYER_REGEN_ENABLED")
	self:UpdatePlayer()
	if not db.playerShowOOC then
		unitFrames.player:Hide()
		unitFrames.pet:Hide()
	end
	if not db.targetShowOOC then
		self:PLAYER_TARGET_CHANGED()
	end
	if db.alphaMode == "COMBAT" then
		local a = db.alphaOOC
		unitFrames.player:SetAlpha(a)
		unitFrames.target:SetAlpha(a)
		unitFrames.pet:SetAlpha(a)
	end
end

------------------------------------------------------------------------

function OnScreenHealth:PLAYER_DEAD()
	Debug(1, "PLAYER_DEAD")
	self:UpdateCombatStatus(playerUnit)
	unitFrames.player:Hide()
	unitFrames.pet:Hide()
end

function OnScreenHealth:PLAYER_ALIVE()
	Debug(1, "PLAYER_ALIVE")
	if not UnitIsGhost(playerUnit) and db.playerShowOOC or UnitAffectingCombat(playerUnit) then
		unitFrames.player:Show()
		self:UNIT_HEALTH(playerUnit)
		if db.showPower then
			self:UNIT_POWER(playerUnit)
		end
		self:UNIT_PET("player")
	end
end

OnScreenHealth.PLAYER_UNGHOST = OnScreenHealth.PLAYER_ALIVE

------------------------------------------------------------------------

function OnScreenHealth:CreateFrames()
	Debug(1, "CreateFrames")
	local f, x, y

	if db.fontShadow then
		x, y = 1, -1
	else
		x, y = 0, 0
	end

	f = CreateFrame("Frame", "OnScreenHealthPlayerFrame", UIParent)
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	f.Combat = f:CreateFontString(nil, "OVERLAY")
	f.Combat:SetShadowOffset(x, y)

	f.Resource = f:CreateFontString(nil, "OVERLAY")
	f.Resource:SetShadowOffset(x, y)

	unitFrames["player"] = f
	unitFrames["vehicle"] = f

	f = CreateFrame("Frame", "OnScreenHealthPetFrame", UIParent)
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	unitFrames["pet"] = f

	f = CreateFrame("Frame", "OnScreenHealthTargetFrame", UIParent)
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	f.Target = f:CreateFontString(nil, "OVERLAY")
	f.Target:SetShadowOffset(x, y)

	f.ComboPoints = f:CreateFontString(nil, "OVERLAY")
	f.ComboPoints:SetShadowOffset(x, y)

	unitFrames["target"] = f

	local a
	if db.alphaMode == "COMBAT" then
		a = UnitAffectingCombat(playerUnit) and db.alphaCombat or db.alphaOOC
	else
		a = db.alphaFixed
	end
	unitFrames["player"]:SetAlpha(a)
	unitFrames["pet"]:SetAlpha(a)
	unitFrames["target"]:SetAlpha(a)

	self:ApplyPositions()
	self:ApplyFonts()
	self:UpdatePlayer()
	self:PLAYER_TARGET_CHANGED()
	self:UNIT_PET("player")
end

------------------------------------------------------------------------

function OnScreenHealth:ApplyPositions()
	Debug(1, "ApplyPositions")

	local p1, p2, p3, p4, p5, p6

	if db.textAlign == "CENTER" then
		p1, p2, p3, p4, p5, p6 = "TOP", "BOTTOM", "CENTER", "TOP", "BOTTOM", "CENTER"
	elseif db.textAlign == "OUTSIDE" then
		p1, p2, p3, p4, p5, p6 = "TOPLEFT", "BOTTOMLEFT", "LEFT", "TOPRIGHT", "BOTTOMRIGHT", "RIGHT"
	else
		p1, p2, p3, p4, p5, p6 = "TOPRIGHT", "BOTTOMRIGHT", "RIGHT", "TOPLEFT", "BOTTOMLEFT", "LEFT"
	end

	local f = unitFrames.player

	f:ClearAllPoints()
	f:SetPoint(p1, UIParent, "CENTER", -db.posX, db.posY)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p1, f, p1, 0, 0)
	f.Health:SetJustifyH(p3)

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p1, f.Health, p2, 0, 0)
	f.Power:SetJustifyH(p3)

	f.Combat:ClearAllPoints()
	f.Combat:SetPoint(p2, f.Power, p1, 0, 0)
	f.Combat:SetJustifyH(p3)

	f.Resource:ClearAllPoints()
	f.Resource:SetPoint(p1, f.Power, p2, 0, 0)
	f.Resource:SetJustifyH(p3)

	f = unitFrames.target

	f:ClearAllPoints()
	f:SetPoint(p4, UIParent, "CENTER", db.posX, db.posY)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p4, f, p4, 0, 0)
	f.Health:SetJustifyH("CENTER")

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p4, f.Health, p5, 0, 0)
	f.Power:SetJustifyH(p6)

	f.Target:ClearAllPoints()
	f.Target:SetPoint(p4, f.Power, p5, 0, 0)
	f.Target:SetJustifyH(p6)

	f.ComboPoints:ClearAllPoints()
	f.ComboPoints:SetPoint(p5, f.Health, p4, 0, 0)
	f.ComboPoints:SetJustifyH(p6)

	f = unitFrames.pet

	f:ClearAllPoints()
	f:SetPoint(p2, unitFrames.player, p1, 0, db.posPet)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p1, f, p1, 0, 0)
	f.Health:SetJustifyH(p3)

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p1, f.Health, p2, 0, 0)
	f.Power:SetJustifyH(p3)
end

------------------------------------------------------------------------

function OnScreenHealth:ApplyFonts()
	Debug(1, "ApplyFonts")

	local face = SharedMedia and SharedMedia:Fetch("font", db.fontFace) or "Fonts\\FRIZQT__.ttf"
	local s1, s2, outline = db.fontSizeHealth, db.fontSizePower, db.fontOutline

	local f = unitFrames.player
	f.Health:SetFont(face, s1, outline)
	f.Power:SetFont(face, s2, outline)
	f.Combat:SetFont(face, s2, outline)
	f.Resource:SetFont(face, s2, outline)

	f = unitFrames.pet
	f.Health:SetFont(face, s1 * db.fontSizePet, outline)
	f.Power:SetFont(face, s2 * db.fontSizePet, outline)

	f = unitFrames.target
	f.Health:SetFont(face, s1, outline)
	f.Power:SetFont(face, s2, outline)
	f.Target:SetFont(face, db.fontSizeTarget, outline)
	f.ComboPoints:SetFont(face, db.fontSizeCombo, outline)
end

------------------------------------------------------------------------

function OnScreenHealth:PLAYER_TALENT_UPDATE()
	if db.showPower and ((playerClass == "SHAMAN" and select(5, GetTalentInfo(2, 17)) > 0) or (playerClass == "PRIEST" and GetPrimaryTalentTree() == 3)) then
		self:RegisterEvent("UNIT_AURA")
		self:UNIT_AURA("player")
	else
		self:UnregisterEvent("UNIT_AURA")
		unitFrames.player.Resource:SetText()
	end
end

function OnScreenHealth:UpdatePowerEvents()
	Debug(1, "UpdatePowerEvents, showPower = %s", tostring(db.showPower))
	if db.showPower then
		self:RegisterEvent("UNIT_POWER")
		self:RegisterEvent("UNIT_MAXPOWER")
		self:RegisterEvent("UNIT_DISPLAYPOWER")
		if (playerClass == "SHAMAN" and select(5, GetTalentInfo(2, 17)) > 0) or (playerClass == "PRIEST" and GetPrimaryTalentTree() == 3) then
			self:RegisterEvent("UNIT_AURA")
		end
	else
		self:UnregisterEvent("UNIT_POWER")
		self:UnregisterEvent("UNIT_MAXPOWER")
		self:UnregisterEvent("UNIT_DISPLAYPOWER")
		self:UnregisterEvent("UNIT_AURA")
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UpdateCombatEvents()
	if db.playerShowOOC and db.targetShowOOC and db.alphaMode ~= "COMBAT" and not db.playerShowCombatStatus then
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	else
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
	end
end

------------------------------------------------------------------------

function OnScreenHealth:UpdateDeathEvents()
	if db.playerShowDead then
		self:UnregisterEvent("PLAYER_ALIVE")
		self:UnregisterEvent("PLAYER_UNGHOST")
		self:UnregisterEvent("PLAYER_DEAD")
	else
		self:RegisterEvent("PLAYER_ALIVE")
		self:RegisterEvent("PLAYER_UNGHOST", "PLAYER_ALIVE")
		self:RegisterEvent("PLAYER_DEAD")
	end
end

------------------------------------------------------------------------

function OnScreenHealth:ADDON_LOADED(addon)
	if addon ~= ADDON_NAME then return end
	Debug(1, "ADDON_LOADED", addon)

	local defaults = {
		playerEnable = true,
		playerShowDead = false,
		playerShowOOC = true,
		playerShowCombatStatus = false,
		playerFrequentUpdates = false,
		targetEnable = true,
		targetShowDead = true,
		targetShowOOC = true,
		targetShowCP = true,
		targetShowToT = false,
		petEnable = true,
		petShowDead = false,
		alphaMode = "FIXED",
		alphaFixed = 1,
		alphaCombat = 1,
		alphaOOC = 0.5,
		colorHealthMode = "GRADIENT",
		showPower = true,
		textAlign = "INSIDE",
		textFormatHealthMonster = "$c",
		textFormatHealthPlayer = "$c",
		textFormatPowerMonster = "$c",
		textFormatPowerPlayer = "$c",
		fontFace = "Friz Quadrata TT",
		fontOutline = "OUTLINE",
		fontShadow = true,
		fontSizeHealth = 32,
		fontSizePower = 24,
		fontSizeTarget = 24,
		fontSizeCombo = 24,
		fontSizePet = 0.8,
		colorHealthMax = { 0, 1, 0 },
		colorHealthMid = { 1, 1, 0 },
		colorHealthMin = { 1, 0, 0 },
		powerColors = {
			MAELSTROM_WEAPON = "cc99ff",
			SHADOW_ORBS = "cc99ff",
		},
		colorAbsent = "999999",
		colorCombatStatus = "ffffff",
		colorCombo = "ff7f00",
		colorTargetSelf = "ffffff",
		colorDefault = "ff00ff", -- Fallback
		posX = 200,
		posY = 0,
		posPet = 20,
		version = VERSION
	}

	for k, v in pairs(PowerBarColor) do
		if type(k) == "string" and type(v) == "table" and v.r then
			defaults.powerColors[k] = ("%02x%02x%02x"):format(v.r * 255, v.g * 255, v.b * 255)
		end
	end
	for k, v in pairs(PowerBarColor.ECLIPSE) do
		defaults.powerColors[k == "positive" and "ECLIPSE_SOLAR" or "ECLIPSE_LUNAR"] = ("%02x%02x%02x"):format(v.r * 255, v.g * 255, v.b * 255)
	end

	self.db = LibStub("AceDB-3.0"):New("OnScreenHealthDB", { profile = defaults }, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "Refresh")
	self.db.RegisterCallback(self, "OnProfileCopied", "Refresh")
	self.db.RegisterCallback(self, "OnProfileReset", "Refresh")
	db = self.db.profile

	self:LoadOptions()

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil

	if IsLoggedIn() then
		self:PLAYER_LOGIN()
	else
		self:RegisterEvent("PLAYER_LOGIN")
	end
end

------------------------------------------------------------------------

function OnScreenHealth:PLAYER_LOGIN()
	Debug(1, "PLAYER_LOGIN")

	if not unitFrames.player then
		self:CreateFrames()
	end

	self:RegisterEvent("UNIT_HEALTH")
	self:RegisterEvent("UNIT_MAXHEALTH")

	self:RegisterEvent("PLAYER_TARGET_CHANGED")

	self:RegisterEvent("UNIT_ENTERED_VEHICLE")
	self:RegisterEvent("UNIT_EXITED_VEHICLE")

	self:RegisterEvent("PLAYER_TALENT_UPDATE")

	self:UpdatePowerEvents()
	self:UpdateCombatEvents()
	self:UpdateDeathEvents()

	self:RegisterEvent("UNIT_PET")

	if db.playerFrequentUpdates then
		unitFrames.player:SetScript("OnUpdate", self.FrequentUpdate)
	end

	if db.targetShowToT then
		unitFrames.target:SetScript("OnUpdate", self.CheckToT)
	end

	if db.targetShowCP then
		Debug(1, "Registering combo point events")
		self:RegisterEvent("UNIT_COMBO_POINTS")
	end

	self:UnregisterEvent("PLAYER_LOGIN")
end

function OnScreenHealth:Refresh()
	Debug(1, "Refresh")
	db = self.db.profile

	self:UpdatePowerEvents()
	self:UpdateCombatEvents()
	self:UpdateDeathEvents()

	self:ApplyFonts()
	self:ApplyPositions()

	self:UpdateAll()
end

------------------------------------------------------------------------

function OnScreenHealth.FrequentUpdate(frame, elapsed)
	return OnScreenHealth:UNIT_POWER(playerUnit)
end

------------------------------------------------------------------------