--[[--------------------------------------------------------------------
	OnScreenHealth
	Minimal text-based HUD for the player, pet, and target units.
	by Phanx < addons@phanx.net >
	Copyright  20072010 Phanx. Some rights reserved. See LICENSE.txt for details.
	http://www.wowinterface.com/downloads/info7470-OnScreenHealth.html
	http://wow.curse.com/downloads/wow-addons/details/onscreenhealth.aspx
----------------------------------------------------------------------]]

local ADDON_NAME, ns = ...

local OnScreenHealth = ns.OnScreenHealth
if not OnScreenHealth then return end

local frames = ns.unitFrames

local function getColor(info)
	local v = OnScreenHealth.db.profile[ info[#info] ]
	if not v then
		return
	elseif type(v) == "table" then
		return unpack(v)
	else
		return tonumber(v:sub(1, 2), 16) / 255, tonumber(v:sub(3, 4), 16) / 255, tonumber(v:sub(5, 6), 16) / 255
	end
end

local function setColor(info, r, g, b)
	local k = info[#info]
	local t = OnScreenHealth.db.profile
	if type(t[k]) == "table" then
		t[k][1] = r
		t[k][2] = g
		t[k][3] = b
	else
		t[k] = ("%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end
	if k:match( "Health" ) then
		OnScreenHealth:UpdateAll("health")
	else
		-- Currently only using this for health and mana colors
		-- Will require changes if used for other color types!
		OnScreenHealth:UpdateAll("power")
	end
end

local function getPowerColor(info)
	local v = OnScreenHealth.db.profile.powerColors[ info[#info] ]
	if not v then return end
	return tonumber(v:sub(1, 2), 16) / 255, tonumber(v:sub(3, 4), 16) / 255, tonumber(v:sub(5, 6), 16) / 255
end

local function setPowerColor(info, r, g, b)
	local k = info[#info]
	local t = OnScreenHealth.db.profile.powerColors
	if not t[k] then return end
	t[k] = ("%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	OnScreenHealth:UpdateAll("power")
end

local function getPos(info)
	return OnScreenHealth.db.profile[ info[#info] ]
end

local function setPos(info, v)
	OnScreenHealth.db.profile[ info[#info] ] = v
	OnScreenHealth:ApplyPositions()
end

function OnScreenHealth:LoadOptions()
	if self.optionsPanel then return end

	local L = ns.L
	local db = self.db.profile
	local Debug = ns.Debug
	local SharedMedia = LibStub("LibSharedMedia-3.0")

	local options = {
		type = "group", childGroups = "tab",
		args = {
			player = {
				order = 1,
				name = L["Player"],
				desc = L["Basic options for the player display."],
				type = "group",
				args = {
					enable = {
						order = 110,
						name = L["Enable"],
						desc = L["Enable the player display."],
						type = "toggle",
						get = function() return db.playerEnable end,
						set = function(info, v)
							db.playerEnable = v
							self:UpdatePlayer()
							self:UNIT_PET("player")
						end
					},
					showDead = {
						order = 120,
						name = L["Show Dead"],
						desc = L["Show the player display while you are dead."],
						type = "toggle",
						get = function() return db.playerShowDead end,
						set = function(info, v)
							db.playerShowDead = v
							self:UpdateDeathEvents()
							self:UpdatePlayer()
							self:UNIT_PET("player")
						end
					},
					showOOC = {
						order = 130,
						name = L["Show OOC"],
						desc = L["Show the player display while you are out of combat."],
						type = "toggle",
						get = function() return db.playerShowOOC end,
						set = function(info, v)
							db.playerShowOOC = v
							self:UpdateCombatEvents()
							self:UpdatePlayer()
							self:UNIT_PET("player")
						end
					},
					showCombatStatus = {
						order = 140,
						name = L["Show Combat Status"],
						desc = L["Show an indicator above player health while you are in combat."],
						type = "toggle",
						get = function() return db.playerShowCombatStatus end,
						set = function(info, v)
							db.playerShowCombatStatus = v
							self:UpdateCombatEvents()
							self:UpdatePlayer()
						end,
					},
					frequentUpdates = {
						order = 150,
						name = L["Frequent Power Updates"],
						desc = L["Enable updating the power text more frequently than normal."],
						type = "toggle",
						get = function() return db.playerFrequentUpdates end,
						set = function(info, v)
							db.playerFrequentUpdates = v
							if v then
								frames.player:SetScript("OnUpdate", OnScreenHealth.FrequentUpdate)
							else
								frames.player:SetScript("OnUpdate", nil)
							end
						end
					}
				}
			},
			pet = {
				order = 2,
				name = L["Pet"],
				desc = L["Basic options for the pet display."],
				type = "group",
				args = {
					enable = {
						order = 210,
						name = L["Enable"],
						desc = L["Enable the pet display."],
						type = "toggle",
						get = function() return db.petEnable end,
						set = function(info, v)
							db.petEnable = v
							self:UNIT_PET("player")
						end,
					},
					showDead = {
						order = 220,
						name = L["Show Dead"],
						desc = L["Show the pet display while your pet is dead."],
						type = "toggle",
						get = function() return db.petShowDead end,
						set = function(info, v)
							db.petShowDead = v
							self:UNIT_PET("player")
						end
					}
				}
			},
			target = {
				order = 3,
				name = L["Target"],
				desc = L["Basic options for the target display."],
				type = "group",
				args = {
					enable = {
						order = 310,
						name = L["Enable"],
						desc = L["Enable the target display."],
						type = "toggle",
						get = function() return db.targetEnable end,
						set = function(info, v)
							db.targetEnable = v
							self:PLAYER_TARGET_CHANGED()
						end
					},
					showDead = {
						order = 320,
						name = L["Show Dead"],
						desc = L["Show the target display while your target is dead or offline."],
						type = "toggle",
						get = function() return db.targetShowDead end,
						set = function(info, v)
							db.targetShowDead = v
							self:PLAYER_TARGET_CHANGED()
						end
					},
					showOOC = {
						order = 330,
						name = L["Show OOC"],
						desc = L["Show the target display while you are out of combat."],
						type = "toggle",
						get = function() return db.targetShowOOC end,
						set = function(info, v)
							db.targetShowOOC = v
							self:UpdateCombatEvents()
							self:PLAYER_TARGET_CHANGED()
						end
					},
					showCP = {
						order = 340,
						name = L["Combo Points"],
						desc = L["Show your combo points."],
						type = "toggle",
						get = function() return db.targetShowCP end,
						set = function(info, v)
							db.targetShowCP = v
							if v then
								self:RegisterEvent("UNIT_COMBO_POINTS", "UpdateCP")
								self:UNIT_COMBO_POINTS()
							else
								self:UnregisterEvent("UNIT_COMBO_POINTS")
								frames.target.ComboPoints:SetText()
							end
						end
					},
					showToT = {
						order = 350,
						name = L["Target of Target"],
						desc = L["Show the name of your target's target."],
						type = "toggle",
						get = function() return db.targetShowToT end,
						set = function(info, v)
							db.targetShowToT = v
							if v then
								frames.target:SetScript("OnUpdate", OnScreenHealth.CheckToT)
							else
								frames.target:SetScript("OnUpdate", nil)
							end
						end
					}
				}
			},
			appearance = {
				order = 4,
				name = L["Appearance"],
				desc = L["Options related to general appearance."],
				type = "group",
				args = {
					alpha = {
						name = L["Alpha"],
						desc = L["Options related to transparency."],
						type = "group", dialogInline = true,
						args = {
							mode = {
								order = 611,
								name = L["Mode"],
								desc = L["Choose how transparency should be set."],
								type = "select", values = { ["FIXED"] = L["Fixed Alpha"], ["COMBAT"] = L["Combat Fade"], ["HEALTH"] = L["Health Fade"] },
								get = function() return db.alphaMode end,
								set = function(info, v)
									db.alphaMode = v
									self:UpdateCombatEvents()
									if v == "COMBAT" then
										local a = UnitAffectingCombat("player") and db.alphaCombat or db.alphaOOC
										frames.player:SetAlpha(a)
										frames.target:SetAlpha(a)
										frames.pet:SetAlpha(a)
									else
										if v == "HEALTH" then
											self:UNIT_HEALTH("player")
											self:UNIT_HEALTH("target")
											self:UNIT_HEALTH("pet")
										else
											local a = db.alphaFixed
											frames.player:SetAlpha(a)
											frames.target:SetAlpha(a)
											frames.pet:SetAlpha(a)
										end
									end
								end,
							},
							fixed = {
								order = 612,
								name = L["Alpha"],
								desc = L["Set a fixed transparency for the displays."],
								hidden = function() return db.alphaMode ~= "FIXED" end,
								type = "range", min = 0.05, max = 1.00, step = 0.01, bigStep = 0.05, isPercent = true,
								get = function() return db.alphaFixed end,
								set = function(info, v)
									db.alphaFixed = v
									if db.alphaMode == "FIXED" then
										frames.player:SetAlpha(v)
										frames.target:SetAlpha(v)
										frames.pet:SetAlpha(v)
									end
								end,
							},
							combat = {
								order = 613,
								name = L["Alpha In Combat"],
								desc = L["Set the transparency level to use in combat."],
								hidden = function() return db.alphaMode ~= "COMBAT" end,
								type = "range", min = 0.05, max = 1.00, step = 0.01, bigStep = 0.05, isPercent = true,
								get = function() return db.alphaCombat end,
								set = function(info, v)
									db.alphaCombat = v
									if db.alphaMode == "COMBAT" and UnitAffectingCombat("player") then
										frames.player:SetAlpha(v)
										frames.target:SetAlpha(v)
										frames.pet:SetAlpha(v)
									end
								end,
							},
							ooc = {
								order = 614,
								name = L["Alpha Out Of Combat"],
								desc = L["Set the transparency level to use out of combat."],
								hidden = function() return db.alphaMode ~= "COMBAT" end,
								type = "range", min = 0.05, max = 1.00, step = 0.01, bigStep = 0.05, isPercent = true,
								get = function() return db.alphaOOC end,
								set = function(info, v)
									db.alphaOOC = v
									if db.alphaMode == "COMBAT" and not UnitAffectingCombat("player") then
										frames.player:SetAlpha(v)
										frames.target:SetAlpha(v)
										frames.pet:SetAlpha(v)
									end
								end,
							},
						}
					},
					colorMode = {
						name = L["Health Color Mode"],
						desc = L["Choose how to color the unit's health."],
						type = "select", values = { ["FIXED"] = L["Solid Color Thresholds"], ["GRADIENT"] = L["Smooth Color Gradient"], ["CLASS"] = L["Player Class"] },
						get = function() return db.colorHealthMode end,
						set = function(info, v)
							db.colorHealthMode = v
							self:UpdateAll("health")
						end
					},
					showPower = {
						name = L["Show Power Text"],
						desc = L["Show text for mana, rage, energy, and focus."],
						type = "toggle",
						get = function() return db.showPower end,
						set = function(info, v)
							db.showPower = v
							self:UpdatePowerEvents()
							if v then
								self:UNIT_POWER("player")
								self:UNIT_POWER("target")
								self:UNIT_POWER("pet")
							else
								frames.player:SetScript("OnUpdate", nil)
								frames.player.Power:SetText()
								frames.target.Power:SetText()
								frames.pet.Power:SetText()
							end
						end
					},
					textAlign = {
						name = L["Text Alignment"],
						desc = L["Set the text alignment."],
						type = "select", values = { ["CENTER"] = L["Center"], ["INSIDE"] = L["Inside"], ["OUTSIDE"] = L["Outside"] },
						get = function() return db.textAlign end,
						set = function(info, v)
							db.textAlign = v
							self:ApplyPositions()
						end
					},
					textFormat = {
						name = L["Text Format"],
						desc = L["Set the format for displayed text."],
						type = "group", dialogInline = true,
						args = {
							health = {
								name = L["Health Format"],
								desc = L["Set the formats for health text."],
								type = "group", dialogInline = true,
								get = function(info) return db[ info[#info] ] end,
								set = function(info, v)
									db[ info[#info] ] = v
									self:UpdateAll("health")
								end,
								args = {
									textFormatHealthMonster = {
										name = L["NPC Health"],
										desc = L["Set the format for NPCs' health text."]..L["\n$p = percent \n$c = current \n$m = maximum \n$d = missing \n$c:s = short current \n$m:s = short maximum \n$d:s = short missing"],
										type = "input",
										usage = "Enter a text format for NPC health.",
									},
									textFormatHealthPlayer = {
										name = L["Player Health"],
										desc = L["Set the format for players' health text."]..L["\n$p = percent \n$c = current \n$m = maximum \n$d = missing \n$c:s = short current \n$m:s = short maximum \n$d:s = short missing"],
										type = "input",
										usage = "Enter a text format for player health.",
									}
								}
							},
							power = {
								name = L["Power Format"],
								desc = L["Set the formats for power text."],
								type = "group", dialogInline = true,
								get = function(info) return db[ info[#info] ] end,
								set = function(info, v)
									db[ info[#info] ] = v
									self:UpdateAll("power")
								end,
								args = {
									textFormatPowerMonster = {
										name = L["NPC Power"],
										desc = L["Set the format for NPCs' power text."]..L["\n$p = percent \n$c = current \n$m = maximum \n$d = missing \n$c:s = short current \n$m:s = short maximum \n$d:s = short missing"],
										type = "input",
										usage = "Enter a text format for NPC power.",
									},
									textFormatPowerPlayer = {
										name = L["Player Power"],
										desc = L["Set the format for players' power text."]..L["\n$p = percent \n$c = current \n$m = maximum \n$d = missing \n$c:s = short current \n$m:s = short maximum \n$d:s = short missing"],
										type = "input",
										usage = "Enter a text format for player power.",
									}
								}
							}
						}
					}
				}
			},
			font = {
				order = 5,
				name = L["Font"],
				desc = L["Options relating to the font used."],
				type = "group",
				args = {
					face = {
						order = 710,
						name = L["Typeface"],
						desc = L["Set the typeface to be used."],
						type = "select", values = AceGUIWidgetLSMlists.font, dialogControl = "LSM30_Font",
						get = function() return db.fontFace end,
						set = function(info, v)
							db.fontFace =  v
							self:ApplyFonts()
						end
					},
					outline = {
						order = 720,
						name = L["Outline"],
						desc = L["Set the text outline to be used."],
						type = "select", values = { ["NONE"] = L["None"], ["OUTLINE"] = L["Thin"], ["THICKOUTLINE"] = L["Thick"] },
						get = function() return db.fontOutline end,
						set = function(info, v)
							db.fontOutline = v
							self:ApplyFonts()
						end
					},
					shadow = {
						order = 730,
						name = L["Shadow"],
						desc = L["Enable a text shadow."],
						type = "toggle",
						get = function() return db.fontShadow end,
						set = function(info, v)
							db.fontShadow = v

							local s1, s2
							if v then
								s1, s2 = 2, -2
							else
								s1, s2 = 0, 0
							end

							frames.player.Health:SetShadowOffset(s1, s2)
							frames.player.Power:SetShadowOffset(s1, s2)
							frames.player.Combat:SetShadowOffset(s1, s2)
							frames.player.Resource:SetShadowOffset(s1, s2)

							frames.target.Health:SetShadowOffset(s1, s2)
							frames.target.Power:SetShadowOffset(s1, s2)
							frames.target.Target:SetShadowOffset(s1, s2)
							frames.target.ComboPoints:SetShadowOffset(s1, s2)

							frames.pet.Health:SetShadowOffset(s1, s2)
							frames.pet.Power:SetShadowOffset(s1, s2)
						end
					},
					size = {
						order = 740,
						name = L["Size"],
						desc = L["Options relating to font size."],
						type = "group", dialogInline = true,
						args = {
							health = {
								order = 741,
								name = L["Health"],
								desc = L["Set the font size used for health text."],
								type = "range", min = 8, max = 32, step = 1,
								get = function() return db.fontSizeHealth end,
								set = function(info, v)
									db.fontSizeHealth = v
									local face, outline = SharedMedia:Fetch("font", db.fontFace), db.fontOutline
									frames.player.Health:SetFont(face, v, outline)
									frames.target.Health:SetFont(face, v, outline)
									frames.pet.Health:SetFont(face, db.fontSizePet * v, outline)
								end
							},
							power = {
								order = 742,
								name = L["Power"],
								desc = L["Set the font size used for power text."],
								type = "range", min = 8, max = 32, step = 1,
								get = function() return db.fontSizePower end,
								set = function(info, v)
									db.fontSizePower = v
									local face, outline = SharedMedia:Fetch("font", db.fontFace), db.fontOutline
									frames.player.Power:SetFont(face, v, outline)
									frames.target.Power:SetFont(face, v, outline)
									frames.pet.Power:SetFont(face, db.fontSizePet * v, outline)
								end
							},
							combo = {
								order = 743,
								name = L["Combo Points"],
								desc = L["Set the font size used for combo points."],
								type = "range", min = 8, max = 32, step = 1,
								get = function() return db.fontSizeCombo end,
								set = function(info, v)
									db.fontSizeCombo = v
									frames.player.Combat:SetFont(SharedMedia:Fetch("font", db.fontFace), v, db.fontOutline)
									frames.target.ComboPoints:SetFont(SharedMedia:Fetch("font", db.fontFace), v, db.fontOutline)
								end
							},
							target = {
								order = 744,
								name = L["Target Name"],
								desc = L["Set the font size used for target name text."],
								type = "range", min = 8, max = 32, step = 1,
								get = function() return db.fontSizeTarget end,
								set = function(info, v)
									db.fontSizeTarget = v
									frames.player.Resource:SetFont(SharedMedia:Fetch("font", db.fontFace), v, db.fontOutline)
									frames.target.Target:SetFont(SharedMedia:Fetch("font", db.fontFace), v, db.fontOutline)
								end
							},
							pet = {
								order = 745,
								name = L["Pet Scale"],
								desc = L["Set the scale for pet health and power text compared to player text."],
								type = "range", min = 0.1, max = 1.0, step = 0.1, isPercent = true,
								get = function() return db.fontSizePet end,
								set = function(info, v)
									db.fontSizePet = v
									local face, outline = SharedMedia:Fetch("font", db.fontFace), db.fontOutline
									frames.pet.Health:SetFont(face, db.fontSizeHealth * v, outline)
									frames.pet.Power:SetFont(face, db.fontSizePower * v, outline)
								end
							}
						}
					}
				}
			},
			colors = {
				order = 6,
				name = L["Colors"],
				type = "group",
				args = {
					health = {
						name = L["Health"],
						type = "group", dialogInline = true,
						get = getColor,
						set = setColor,
						args = {
							colorHealthMax = {
								order = 811,
								name = L["Full Health"],
								desc = string.format(L["Set the color used for %s."], L["Full Health"]),
								type = "color",
							},
							colorHealthMid = {
								order = 812,
								name = L["Medium Health"],
								desc = string.format(L["Set the color used for %s."], L["Medium Health"]),
								type = "color",
							},
							colorHealthMin = {
								order = 813,
								name = L["Low Health"],
								desc = string.format(L["Set the color used for %s."], L["Low Health"]),
								type = "color",
							}
						}
					},
					power = {
						name = L["Power"],
						type = "group", dialogInline = true,
						get = getPowerColor,
						set = setPowerColor,
						args = {
							MANA = {
								order = 821,
								name = L["Mana"],
								desc = string.format(L["Set the color used for %s."], L["Mana"]),
								type = "color",
							},
							RAGE = {
								order = 822,
								name = L["Rage"],
								desc = string.format(L["Set the color used for %s."], L["Rage"]),
								type = "color",
							},
							ENERGY = {
								order = 823,
								name = L["Energy"],
								desc = string.format(L["Set the color used for %s."], L["Energy"]),
								type = "color",
							},
							FOCUS = {
								order = 824,
								name = L["Focus"],
								desc = string.format(L["Set the color used for %s."], L["Focus"]),
								type = "color",
							},
							RUNIC_POWER = {
								order = 825,
								name = L["Runic Power"],
								desc = string.format(L["Set the color used for %s."], L["Runic Power"]),
								type = "color",
							},
							HOLY_POWER = {
								order = 826,
								name = L["Holy Power"],
								desc = string.format(L["Set the color used for %s."], L["Holy Power"]),
								type = "color",
							},
							SOUL_SHARDS = {
								order = 827,
								name = L["Soul Shards"],
								desc = string.format(L["Set the color used for %s."], L["Soul Shards"]),
								type = "color",
							},
							MAELSTROM_WEAPON = {
								order = 828,
								name = L["Maelstrom Weapon"],
								desc = string.format(L["Set the color used for %s."], L["Maelstrom Weapon"]),
								type = "color",
							},
							SHADOW_ORBS = {
								order = 829,
								name = L["Shadow Orbs"],
								desc = string.format(L["Set the color used for %s."], L["Shadow Orbs"]),
								type = "color",
							},
						}
					},
					absent = {
						name = L["Absent"],
						desc = string.format(L["Set the color used for %s."], L["dead and offline units"]),
						type = "color",
						get = function()
							local v = db.colorAbsent
							return tonumber(v:sub(1, 2), 16) / 255, tonumber(v:sub(3, 4), 16) / 255, tonumber(v:sub(5, 6), 16) / 255
						end,
						set = function(info, r, g, b)
							db.colorAbsent = string.format("%02x%02x%02x", r*255, g*255, b*255)
							self:UNIT_HEALTH("player")
							self:UNIT_HEALTH("target")
							self:UNIT_HEALTH("pet")
						end
					},
					combatStatus = {
						name = L["Combat Status"],
						desc = string.format(L["Set the color used for %s"], L["the player combat status indicator."]),
						type = "color",
						get = function()
							local tmp1 = db.colorCombatStatus
							return tonumber(tmp1:sub(1, 2), 16) / 255, tonumber(tmp1:sub(3, 4), 16) / 255, tonumber(tmp1:sub(5, 6), 16) / 255
						end,
						set = function(info, r, g, b)
							db.colorCombatStatus = string.format("%02x%02x%02x", r*255, g*255, b*255)
							self:UpdateCombatStatus("player")
						end
					},
					combo = {
						name = L["Combo Points"],
						desc = string.format(L["Set the color used for %s."], L["Combo Points"]),
						type = "color",
						get = function()
							local tmp1 = db.colorCombo
							return tonumber(tmp1:sub(1, 2), 16) / 255, tonumber(tmp1:sub(3, 4), 16) / 255, tonumber(tmp1:sub(5, 6), 16) / 255
						end,
						set = function(info, r, g, b)
							db.colorCombo = string.format("%02x%02x%02x", r*255, g*255, b*255)
							self:UNIT_COMBO_POINTS()
						end
					},
					targetSelf = {
						name = L["Target Self"],
						desc = string.format(L["Set the color used for %s."], L["your name when you are your target's target"]),
						type = "color",
						get = function()
							local tmp1 = db.colorTargetSelf
							return tonumber(tmp1:sub(1, 2), 16) / 255, tonumber(tmp1:sub(3, 4), 16) / 255, tonumber(tmp1:sub(5, 6), 16) / 255
						end,
						set = function(info, r, g, b)
							db.colorTargetSelf = string.format("%02x%02x%02x", r*255, g*255, b*255)
							self:UpdateToT()
						end
					}
				}
			},
			position = {
				order = 7,
				name = L["Position"],
				desc = L["Set the position of the displays on-screen."],
				type = "group",
				get = getPos,
				set = setPos,
				args = {
					posX = {
						order = 910,
						name = L["Horizontal"],
						desc = L["Set the horizontal distance from the center of the screen."],
						type = "range", min = 0, max = 800, step = 10, bigStep = 50,
					},
					posY = {
						order = 920,
						name = L["Vertical"],
						desc = L["Set the vertical distance from the center of the screen."],
						type = "range", min = -500, max = 500, step = 10, bigStep = 25,
					},
					posPet = {
						order = 930,
						name = L["Pet Offset"],
						desc = L["Set the distance to position the pet frame from the player frame."],
						type = "range", min = -200, max = 200, step = 10, bigStep = 25,
					}
				}
			}
		}
	}

	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	options.args.profile.order = 8

	self.options = LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("OnScreenHealth", options)

	self.optionsPanel = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("OnScreenHealth")
	self.aboutPanel = LibStub("LibAboutPanel").new("OnScreenHealth", "OnScreenHealth")

	LibStub("AceConsole-3.0"):RegisterChatCommand("osh", function()
		InterfaceOptionsFrame_OpenToCategory(OnScreenHealth.aboutPanel)
		InterfaceOptionsFrame_OpenToCategory(OnScreenHealth.optionsPanel)
		InterfaceOptionsFrame:Raise()
	end)
end